@@ -30,6 +30,7 @@
 #include <time.h>
 #include <ctype.h>
 #include <errno.h>
+#include <libgen.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -44,9 +45,15 @@
 #include <libaudit.h>
 #endif
 
+#if HAVE_SYSTEMD
+#include <systemd/sd-bus.h>
+#define _cleanup_(f) __attribute__((cleanup(f)))
+#endif
+
+#include "import.h"
 #include "wtmpdb.h"
 
-static char *wtmpdb_path = _PATH_WTMPDB;
+static char *wtmpdb_path = NULL;
 
 #define TIMEFMT_CTIME  1
 #define TIMEFMT_SHORT  2
@@ -58,6 +65,9 @@ static char *wtmpdb_path = _PATH_WTMPDB;
 
 #define LOGROTATE_DAYS 60
 
+/* lenght of login string cannot become longer */
+#define LAST_TIMESTAMP_LEN 32
+
 static uint64_t wtmp_start = UINT64_MAX;
 static int after_reboot = 0;
 
@@ -67,6 +77,7 @@ static int nohostname = 0;
 static int noservice = 1;
 static int dflag = 0;
 static int iflag = 0;
+static int jflag = 0;
 static int wflag = 0;
 static int xflag = 0;
 static const int name_len = 8; /* LAST_LOGIN_LEN */
@@ -123,35 +134,56 @@ isipaddr (const char *string, int *addr_type,
 static int
 parse_time (const char *str, time_t *arg)
 {
+  const char *abs_datetime_fmts[] = {
+    "%Y%m%d%H%M%S",
+    "%Y-%m-%d %T",
+    "%Y-%m-%d %R",
+    "%Y-%m-%d",
+    NULL
+  };
+  const char *abs_time_fmts[] = {
+    "%T",
+    "%R",
+    NULL
+  };
+  const char **fmt;
+  struct tm rst = { .tm_isdst = -1 };
   struct tm res;
+  char *r = NULL;
+  time_t t = time (NULL);
 
-  if (strcmp (str, "today") == 0)
+  for (fmt = abs_datetime_fmts; *fmt; fmt++)
     {
-      time_t t = time (NULL);
-      localtime_r (&t, &res);
-      res.tm_isdst = -1;
-      res.tm_sec = res.tm_min = res.tm_hour = 0;
-    }
-  else if (strcmp (str, "yesterday") == 0)
-    {
-      time_t t = time (NULL);
-      localtime_r (&t, &res);
-      res.tm_isdst = -1;
-      res.tm_mday--;
-      res.tm_sec = res.tm_min = res.tm_hour = 0;
-    }
-  else
-    {
-      char *r = strptime (str, "%Y-%m-%d %T",  &res);
-
-      if (r == NULL || *r != '\0')
-	{
-	  r = strptime (str, "%Y-%m-%d",  &res);
-	  if (r == NULL || *r != '\0')
-	    return -1;
-	}
+      res = rst;
+      r = strptime (str, *fmt, &res);
+      if (r != NULL && *r == '\0') goto done;
     }
 
+  /* Use today's date for time-only specs */
+  localtime_r (&t, &rst);
+  rst.tm_isdst = -1;
+
+  for (fmt = abs_time_fmts; *fmt; fmt++)
+    {
+      res = rst;
+      r = strptime (str, *fmt, &res);
+      if (r != NULL && *r == '\0') goto done;
+    }
+
+  if (strcmp (str, "now") == 0)
+    goto done;
+
+  /* Reset time of day for date-only specs */
+  res.tm_sec = res.tm_min = res.tm_hour = 0;
+
+  if (strcmp (str, "yesterday") == 0)
+    res.tm_mday--;
+  else if (strcmp (str, "tomorrow") == 0)
+    res.tm_mday++;
+  else if (strcmp (str, "today") != 0)
+    return -1;
+
+done:
   *arg = mktime (&res);
 
   return 0;
@@ -253,45 +285,82 @@ calc_time_length(char *dst, size_t dstlen, uint64_t start, uint64_t stop)
     snprintf (dst, dstlen, " (00:%02d)", mins);
 }
 
+/* map "soft-reboot" to "s-reboot" if we have only 8 characters
+   for user output (no -w specified) */
+static const char *
+map_soft_reboot (const char *user)
+{
+  if (wflag || strcmp (user, "soft-reboot") != 0)
+    return user;
+
+  if ((int)strlen (user) > name_len)
+    return "s-reboot";
+
+  return user;
+}
+
+static const char *
+remove_parentheses(const char *str)
+{
+
+  static char buf[LAST_TIMESTAMP_LEN];
+
+  if (strlen(str) >= LAST_TIMESTAMP_LEN)
+    return str;
+
+  char *cp = strchr (str, '(');
+
+  if (cp == NULL)
+    return str;
+
+  cp++;
+  strncpy(buf, cp, LAST_TIMESTAMP_LEN);
+
+  cp = strchr (buf, ')');
+  if (cp)
+    *cp = '\0';
+
+  return buf;
+}
+
+static int first_entry = 1;
 static void
 print_line (const char *user, const char *tty, const char *host,
 	    const char *print_service,
 	    const char *logintime, const char *logouttime,
 	    const char *length)
 {
-  char *line;
-
-  if (nohostname)
+  if (jflag)
     {
-      if (asprintf (&line, "%-8.*s %-12.12s%s %-*.*s - %-*.*s %s\n",
-		    wflag?(int)strlen(user):name_len, user, tty, print_service,
-		    login_len, login_len, logintime,
-		    logout_len, logout_len, logouttime,
-		    length) < 0)
+      if (first_entry)
+	first_entry = 0;
+      else
+	printf (",\n");
+      printf ("     { \"user\": \"%s\",\n", user);
+      printf ("       \"tty\": \"%s\",\n", tty);
+      if (!nohostname)
+	printf ("       \"hostname\": \"%s\",\n", host);
+      if (print_service && strlen (print_service) > 0)
+	printf ("       \"service\": \"%s\",\n", print_service);
+      printf ("       \"login\": \"%s\",\n", logintime);
+      if (length[0] == ' ' || length[0] == '(')
 	{
-	  fprintf (stderr, "Out f memory");
-	  exit (EXIT_FAILURE);
+	  printf ("       \"logout\": \"%s\",\n", logouttime);
+	  printf ("       \"length\": \"%s\"\n",  remove_parentheses(length));
 	}
+      else
+	printf ("       \"logout\": \"%s %s\"\n", logouttime, length);
+      printf ("     }");
     }
   else
     {
-      if (hostlast)
+      char *line;
+
+      if (nohostname)
 	{
-	  if (asprintf (&line, "%-8.*s %-12.12s%s %-*.*s - %-*.*s %-12.12s %s\n",
-			wflag?(int)strlen(user):name_len, user, tty, print_service,
-			login_len, login_len, logintime,
-			logout_len, logout_len, logouttime,
-			length, host) < 0)
-	    {
-	      fprintf (stderr, "Out f memory");
-	      exit (EXIT_FAILURE);
-	    }
-	}
-      else
-	{
-	  if (asprintf (&line, "%-8.*s %-12.12s %-16.*s%s %-*.*s - %-*.*s %s\n",
-			wflag?(int)strlen(user):name_len, user, tty,
-			wflag?(int)strlen(host):host_len, host, print_service,
+	  if (asprintf (&line, "%-8.*s %-12.12s%s %-*.*s - %-*.*s %s\n",
+			wflag?(int)strlen (user):name_len,
+			map_soft_reboot (user), tty, print_service,
 			login_len, login_len, logintime,
 			logout_len, logout_len, logouttime,
 			length) < 0)
@@ -300,10 +369,39 @@ print_line (const char *user, const char *tty, const char *host,
 	      exit (EXIT_FAILURE);
 	    }
 	}
-    }
+      else
+	{
+	  if (hostlast)
+	    {
+	      if (asprintf (&line, "%-8.*s %-12.12s%s %-*.*s - %-*.*s %-12.12s %s\n",
+			    wflag?(int)strlen(user):name_len, map_soft_reboot (user),
+			    tty, print_service,
+			    login_len, login_len, logintime,
+			    logout_len, logout_len, logouttime,
+			    length, host) < 0)
+		{
+		  fprintf (stderr, "Out f memory");
+		  exit (EXIT_FAILURE);
+		}
+	    }
+	  else
+	    {
+	      if (asprintf (&line, "%-8.*s %-12.12s %-16.*s%s %-*.*s - %-*.*s %s\n",
+			    wflag?(int)strlen(user):name_len, map_soft_reboot (user), tty,
+			    wflag?(int)strlen(host):host_len, host, print_service,
+			    login_len, login_len, logintime,
+			    logout_len, logout_len, logouttime,
+			    length) < 0)
+		{
+		  fprintf (stderr, "Out f memory");
+		  exit (EXIT_FAILURE);
+		}
+	    }
+	}
 
-  printf ("%s", line);
-  free (line);
+      printf ("%s", line);
+      free (line);
+    }
 }
 
 static int
@@ -340,8 +438,8 @@ print_entry (void *unused __attribute__((__unused__)),
   const char *host = argv[6]?argv[6]:"";
   const char *service = argv[7]?argv[7]:"";
 
-  uint64_t login_t = strtoul(argv[3], &endptr, 10);
-  if ((errno == ERANGE && login_t == UINT64_MAX)
+  uint64_t login_t = strtoull(argv[3], &endptr, 10);
+  if ((errno == ERANGE && login_t == ULLONG_MAX)
       || (endptr == argv[3]) || (*endptr != '\0'))
     fprintf (stderr, "Invalid numeric time entry for 'login': '%s'\n",
 	     argv[3]);
@@ -377,8 +475,8 @@ print_entry (void *unused __attribute__((__unused__)),
 
   if (argv[4])
     {
-      logout_t = strtoul(argv[4], &endptr, 10);
-      if ((errno == ERANGE && logout_t == INT64_MAX)
+      logout_t = strtoull(argv[4], &endptr, 10);
+      if ((errno == ERANGE && logout_t == ULLONG_MAX)
 	  || (endptr == argv[4]) || (*endptr != '\0'))
 	fprintf (stderr, "Invalid numeric time entry for 'logout': '%s'\n",
 		 argv[4]);
@@ -526,14 +624,15 @@ usage (int retval)
   FILE *output = (retval != EXIT_SUCCESS) ? stderr : stdout;
 
   fprintf (output, "Usage: wtmpdb [command] [options]\n");
-  fputs ("Commands: last, boot, boottime, rotate, shutdown\n\n", output);
+  fputs ("Commands: last, boot, boottime, rotate, shutdown, import\n\n", output);
   fputs ("Options for last:\n", output);
   fputs ("  -a, --hostlast      Display hostnames as last entry\n", output);
   fputs ("  -d, --dns           Translate IP addresses into a hostname\n", output);
   fputs ("  -f, --file FILE     Use FILE as wtmpdb database\n", output);
   fputs ("  -F, --fulltimes     Display full times and dates\n", output);
   fputs ("  -i, --ip            Translate hostnames to IP addresses\n", output);
-  fputs ("  -n, --limit N       Display only first N entries\n", output);
+  fputs ("  -j, --json          Generate JSON output\n", output);
+  fputs ("  -n, --limit N, -N   Display only first N entries\n", output);
   fputs ("  -p, --present TIME  Display who was present at TIME\n", output);
   fputs ("  -R, --nohostname    Don't display hostname\n", output);
   fputs ("  -S, --service       Display PAM service used to login\n", output);
@@ -566,6 +665,11 @@ usage (int retval)
   fputs ("  -f, --file FILE     Use FILE as wtmpdb database\n", output);
   fputs ("\n", output);
 
+  fputs ("Options for import (imports legacy wtmp logs):\n", output);
+  fputs ("  -f, --file FILE     Use FILE as wtmpdb database\n", output);
+  fputs ("  logs...             Legacy log files to import\n", output);
+  fputs ("\n", output);
+
   fputs ("Generic options:\n", output);
   fputs ("  -h, --help          Display this help message and exit\n", output);
   fputs ("  -v, --version       Print version number and exit\n", output);
@@ -651,18 +755,32 @@ main_last (int argc, char **argv)
     {"service", no_argument, NULL, 'S'},
     {"since", required_argument, NULL, 's'},
     {"system", no_argument, NULL, 'x'},
-    {"until", required_argument, NULL, 'u'},
+    {"until", required_argument, NULL, 't'},
     {"time-format", required_argument, NULL, TIMEFMT_VALUE},
+    {"json", no_argument, NULL, 'j'},
     {NULL, 0, NULL, '\0'}
   };
   int time_fmt = TIMEFMT_CTIME;
   char *error = NULL;
   int c;
 
-  while ((c = getopt_long (argc, argv, "adf:Fin:p:RSs:t:wx", longopts, NULL)) != -1)
+  while ((c = getopt_long (argc, argv, "0123456789adf:Fijn:p:RSs:t:wx",
+			   longopts, NULL)) != -1)
     {
       switch (c)
         {
+	case '0':
+	case '1':
+	case '2':
+	case '3':
+	case '4':
+	case '5':
+	case '6':
+	case '7':
+	case '8':
+	case '9':
+	  maxentries = maxentries * 10 + c - '0';
+	  break;
 	case 'a':
 	  hostlast = 1;
 	  break;
@@ -681,6 +799,9 @@ main_last (int argc, char **argv)
 	case 'i':
 	  iflag = 1;
 	  break;
+	case 'j':
+	  jflag = 1;
+	  break;
 	case 'n':
 	  maxentries = strtoul (optarg, NULL, 10);
 	  break;
@@ -704,7 +825,7 @@ main_last (int argc, char **argv)
 	case 'S':
 	  noservice = 0;
 	  break;
-	case 'u':
+	case 't':
 	  if (parse_time (optarg, &until) < 0)
 	    {
 	      fprintf (stderr, "Invalid time value '%s'\n", optarg);
@@ -758,6 +879,9 @@ main_last (int argc, char **argv)
       usage (EXIT_FAILURE);
     }
 
+  if (jflag)
+    printf ("{\n   \"entries\": [\n");
+
   if (wtmpdb_read_all (wtmpdb_path, print_entry, &error) != 0)
     {
       if (error)
@@ -772,14 +896,25 @@ main_last (int argc, char **argv)
     }
 
   if (wtmp_start == UINT64_MAX)
-    printf ("%s has no entries\n", wtmpdb_path);
+    {
+      if (!jflag)
+	printf ("%s has no entries\n", wtmpdb_path?wtmpdb_path:"wtmpdb");
+    }
   else if (time_fmt != TIMEFMT_NOTIME)
     {
       char wtmptime[32];
       format_time (time_fmt, wtmptime, sizeof (wtmptime),
 		   wtmp_start/USEC_PER_SEC);
-      printf ("\n%s begins %s\n", wtmpdb_path, wtmptime);
+      if (jflag)
+	printf ("\n   ],\n   \"start\": \"%s\"\n", wtmptime);
+      else
+	printf ("\n%s begins %s\n", wtmpdb_path?wtmpdb_path:"wtmpdb", wtmptime);
     }
+  else if (jflag)
+    printf ("\n   ]\n");
+
+  if (jflag)
+    printf ("}\n");
   return EXIT_SUCCESS;
 }
 
@@ -815,6 +950,51 @@ diff_timespec(const struct timespec *time1, const struct timespec *time0)
   return diff;
 }
 
+#if HAVE_SYSTEMD
+/* Find out if it was a soft-reboot. With systemd v256 we can query systemd
+   for this.
+   Return values:
+   -1: no systemd support
+   0: no soft-reboot
+   >0: number of soft-reboots
+*/
+static int
+soft_reboots_count (void)
+{
+  unsigned soft_reboots_count = -1;
+  _cleanup_(sd_bus_flush_close_unrefp) sd_bus *bus = NULL;
+  sd_bus_error error = SD_BUS_ERROR_NULL;
+  int r;
+
+  if (sd_bus_open_system (&bus) < 0)
+    {
+      fprintf (stderr, "Error: cannot open dbus");
+      return -1;
+    }
+
+  r = sd_bus_get_property_trivial (bus, "org.freedesktop.systemd1",
+				   "/org/freedesktop/systemd1",
+				   "org.freedesktop.systemd1.Manager",
+				   "SoftRebootsCount",
+				   &error, 'u', &soft_reboots_count);
+  if (r < 0)
+    {
+      /* systemd is too old, don't print error */
+      if (!sd_bus_error_has_name (&error, SD_BUS_ERROR_UNKNOWN_PROPERTY))
+	{
+	  /* error occured, log it and return to fallback code */
+	  if (error.message)
+	    fprintf (stderr,
+		     "Failed to get SoftRebootsCount property: %s\n",
+		     error.message);
+	}
+      sd_bus_error_free (&error);
+      return -1;
+    }
+  return soft_reboots_count;
+}
+#endif
+
 static int
 main_boot (int argc, char **argv)
 {
@@ -825,7 +1005,10 @@ main_boot (int argc, char **argv)
   };
   char *error = NULL;
   int c;
+  int soft_reboot = 0;
+#if HAVE_SYSTEMD
   int quiet = 0;
+#endif
 
   while ((c = getopt_long (argc, argv, "f:q", longopts, NULL)) != -1)
     {
@@ -835,7 +1018,9 @@ main_boot (int argc, char **argv)
           wtmpdb_path = optarg;
           break;
 	case 'q':
+#if HAVE_SYSTEMD
 	  quiet = 1;
+#endif
 	  break;
         default:
           usage (EXIT_FAILURE);
@@ -854,13 +1039,25 @@ main_boot (int argc, char **argv)
 
   struct timespec ts_now;
   struct timespec ts_boot;
-  struct timespec ts_empty = { .tv_sec = 0, .tv_nsec = 0 };
   clock_gettime (CLOCK_REALTIME, &ts_now);
+#ifdef CLOCK_BOOTTIME
   clock_gettime (CLOCK_BOOTTIME, &ts_boot);
+#else
+  ts_boot = ts_now;
+#endif
   uint64_t time = wtmpdb_timespec2usec (diff_timespec(&ts_now, &ts_boot));
+#if HAVE_SYSTEMD
+  struct timespec ts_empty = { .tv_sec = 0, .tv_nsec = 0 };
   uint64_t now = wtmpdb_timespec2usec (diff_timespec(&ts_now, &ts_empty));
 
-  if ((now - time) > 300 * USEC_PER_SEC /* 5 minutes */)
+  int count = soft_reboots_count ();
+
+  if (count > 0)
+    {
+      time = now;
+      soft_reboot = 1;
+    }
+  else if ((count < 0) && ((now - time) > 300 * USEC_PER_SEC) /* 5 minutes */)
     {
       if (!quiet)
 	{
@@ -874,13 +1071,15 @@ main_boot (int argc, char **argv)
 	  printf ("Current time: %s\n", timebuf);
 	}
       time = now;
+      soft_reboot = 1;
     }
+#endif
 
 #if HAVE_AUDIT
   log_audit (AUDIT_SYSTEM_BOOT);
 #endif
 
-  if (wtmpdb_login (wtmpdb_path, BOOT_TIME, "reboot", time, "~", uts.release,
+  if (wtmpdb_login (wtmpdb_path, BOOT_TIME, soft_reboot ? "soft-reboot" : "reboot", time, "~", uts.release,
 		    NULL, &error) < 0)
     {
       if (error)
@@ -1011,6 +1210,41 @@ main_shutdown (int argc, char **argv)
   return EXIT_SUCCESS;
 }
 
+static int
+main_import (int argc, char **argv)
+{
+  struct option const longopts[] = {
+    {"file", required_argument, NULL, 'f'},
+    {NULL, 0, NULL, '\0'}
+  };
+  int c;
+
+  while ((c = getopt_long (argc, argv, "f:", longopts, NULL)) != -1)
+    {
+      switch (c)
+        {
+        case 'f':
+          wtmpdb_path = optarg;
+          break;
+        default:
+          usage (EXIT_FAILURE);
+          break;
+        }
+    }
+
+  if (argc == optind)
+    {
+      fprintf (stderr, "No files specified to import.\n");
+      usage (EXIT_FAILURE);
+    }
+
+  for (; optind < argc; optind++)
+    if (import_wtmp_file (wtmpdb_path, argv[optind]) == -1)
+      return EXIT_FAILURE;
+
+  return EXIT_SUCCESS;
+}
+
 int
 main (int argc, char **argv)
 {
@@ -1035,6 +1269,8 @@ main (int argc, char **argv)
     return main_boottime (--argc, ++argv);
   else if (strcmp (argv[1], "rotate") == 0)
     return main_rotate (--argc, ++argv);
+  else if (strcmp (argv[1], "import") == 0)
+    return main_import (--argc, ++argv);
 
   while ((c = getopt_long (argc, argv, "hv", longopts, NULL)) != -1)
     {
@@ -1044,7 +1280,7 @@ main (int argc, char **argv)
 	  usage (EXIT_SUCCESS);
 	  break;
 	case 'v':
-	  printf ("wtmpdb %s\n", PROJECT_VERSION);
+	  printf ("wtmpdb %s\n", VERSION);
 	  break;
 	default:
 	  usage (EXIT_FAILURE);
